#include "CudaLBFGS/lbfgs.h"
#include "CudaLBFGS/error_checking.h"
#include "CudaLBFGS/timer.h"

#include <iostream>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_runtime_api.h>
#include <device_launch_parameters.h>
#include <device_functions.h>


#define cudaCheckErrors(msg) \
    do { \
        cudaError_t __err = cudaGetLastError(); \
        if (__err != cudaSuccess) { \
           fprintf(stderr, "Fatal error: %s (%s at %s:%d)\n",   \
                msg, cudaGetErrorString(__err), \
                __FILE__, __LINE__); \
           fprintf(stderr, "*** FAILED ***\n"); \
        } \
    } while (0)



namespace gpu_d {
   __device__ float m_a;
   __device__ float m_b;
   __device__ float m_c;
}

// this is a kernel:
//
// pass: d_x: "coordinates" (model),
// return: d_f: target function value for these coordinates
//         d_grad: gradients for these coordinates
__global__
void kernel_combined_f_gradf(const float *d_x, float *d_y, float *d_grad) {
   const float &x = *d_x;
   printf("kernel: d_x %p\n", d_x);
   printf("kernel: x %f\n", x);

   // *d_y = gpu_d::m_a * x * x + gpu_d::m_b * x + gpu_d::m_c;
   // *d_grad = 2.0f * gpu_d::m_a * x + gpu_d::m_b;

   *d_y = 2.0 * x * x + 2.0 * x + 4.0;
   *d_grad = 2.0f * 2.0 * x + 2.0;

   // printf("gpu_d::m_a %f m_b %f m_c %f\n", gpu_d::m_a, gpu_d::m_b, gpu_d::m_c);
   printf("       d_y %f\n", *d_y);
}

class gpu_parabola : public cost_function {
public:
     gpu_parabola(float a, float b, float c) : cost_function(1) {

        CudaSafeCall(cudaGetSymbolAddress((void**)&d_a, gpu_d::m_a));
        CudaSafeCall(cudaGetSymbolAddress((void**)&d_b, gpu_d::m_b));
        CudaSafeCall(cudaGetSymbolAddress((void**)&d_c, gpu_d::m_c));

        CudaSafeCall(cudaMemcpy(d_a, &a, sizeof(float), cudaMemcpyHostToDevice) );
        CudaSafeCall(cudaMemcpy(d_b, &b, sizeof(float), cudaMemcpyHostToDevice) );
        CudaSafeCall(cudaMemcpy(d_c, &c, sizeof(float), cudaMemcpyHostToDevice) );

     }

     gpu_parabola(bool arg) : cost_function(1) {}

     // pass: d_x: "coordinates" (model),
     // return: d_f: target function value for these coordinates
     //         d_grad: gradients for these coordinates
     void f_gradf(const float *d_x, float *d_f, float *d_grad) {
          kernel_combined_f_gradf<<<1,1>>>(d_x, d_f, d_grad);
          cudaCheckErrors("cuda kernel fail kernel_combined_f_gradf()");
     }

private:

   // my pointers to gpu_d values

   float *d_a;
   float *d_b;
   float *d_c;

};


int main(int argc, char **argv) {

   int status = 1;

   // gpu_parabola gp(2,2,4);
   // lbfgs minimizer(gp);

   gpu_parabola gp_2(1);
   lbfgs minimizer_2(gp_2);

   float x = 8.6f;
   float *d_x;
   cudaMalloc(&d_x, sizeof(float));
   cudaMemcpy(d_x, &x, sizeof(float), cudaMemcpyHostToDevice);
   minimizer_2.minimize(d_x);
   cudaMemcpy(&x, d_x, sizeof(float), cudaMemcpyDeviceToHost);
   cudaFree(d_x);
   std::cout << "x (minimized): " << x << std::endl;

   return status;

}
