
#include <iostream>
#include <fstream>
#include <sstream>
#include <iterator>
#include "atoms-container-t.hh"

atoms_container_t::atoms_container_t(const std::string &file_name) {

   std::ifstream f(file_name);
   if (f) {
      std::string line;
      while (std::getline(f, line)) {
	 std::istringstream iss(line);
	 std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
					std::istream_iterator<std::string>());
	 if (parts.size() == 3) {
	    pos_t p(std::stof(parts[0]), std::stof(parts[1]), std::stof(parts[2]));
	    positions.push_back(p);
	 }
      }
   }
}

#ifndef STANDALONE_INPUTS
atoms_container_t::atoms_container_t(mmdb::Manager *mol) {

   bool needs_exporting_kludge = true;

   int imod = 1;
   mmdb::Model *model_p = mol->GetModel(imod);
   if (model_p) {
      int n_chains = model_p->GetNumberOfChains();
      for (int ichain=0; ichain<n_chains; ichain++) {
         mmdb::Chain *chain_p = model_p->GetChain(ichain);
         int nres = chain_p->GetNumberOfResidues();
         for (int ires=0; ires<nres; ires++) {
            mmdb::Residue *residue_p = chain_p->GetResidue(ires);
            int n_atoms = residue_p->GetNumberOfAtoms();
            for (int iat=0; iat<n_atoms; iat++) {
               mmdb::Atom *at = residue_p->GetAtom(iat);
               if (!at->isTer() || needs_exporting_kludge) {
                  pos_t p(at->x, at->y, at->z);
                  positions.push_back(p);
                  atoms.push_back(at);
               }
            }
         }
      }
   }
}
#endif // STANDALONE_INPUTS


void
atoms_container_t::write_atoms(const std::string &file_name) const {

   std::ofstream f(file_name);
   if (f) {
      unsigned int n = positions.size();
      for (std::size_t i=0; i<n; i++) {
         f << positions[i].xyz[0] << "  " << positions[i].xyz[1] << "  " << positions[i].xyz[2] << "\n";
      }
      f.close();
   }
}


#ifndef STANDALONE_INPUTS
void
atoms_container_t::transfer_to(atom_selection_container_t &asc) {

   std::cout << "INFO:: atom selection container had " << asc.n_selected_atoms << " atoms and now "
             << "transfering " << positions.size() << " atoms into that container" << std::endl;

   int idx = 0;
   int imod = 1;
   mmdb::Model *model_p = asc.mol->GetModel(imod);
   if (model_p) {
      int n_chains = model_p->GetNumberOfChains();
      for (int ichain=0; ichain<n_chains; ichain++) {
         mmdb::Chain *chain_p = model_p->GetChain(ichain);
         int nres = chain_p->GetNumberOfResidues();
         for (int ires=0; ires<nres; ires++) {
            mmdb::Residue *residue_p = chain_p->GetResidue(ires);
            int n_atoms = residue_p->GetNumberOfAtoms();
            for (int iat=0; iat<n_atoms; iat++) {
               mmdb::Atom *at = residue_p->GetAtom(iat);
               if (!at->isTer()) {
                  at->x = positions[idx].xyz[0];
                  at->y = positions[idx].xyz[1];
                  at->z = positions[idx].xyz[2];
               }
               idx++;
            }
         }
      }
   }
}
#endif // STANDALONE_INPUTS
