
#ifndef STANDALONE_INPUTS
#include <mmdb2/mmdb_manager.h>
#include <coot/geometry/residue-and-atom-specs.hh>
#endif
#include <vector>

#ifndef STANDALONE_INPUTS
#include "coot-utils/atom-selection-container.hh"
#endif

class atoms_container_t {
public:

   class pos_t {
   public:
      float xyz[3];
      pos_t(const float &x, const float &y, const float &z) {
          xyz[0] = x; xyz[1] = y; xyz[2] = z;
      }
      pos_t operator-(const pos_t &p) const {
         return pos_t(xyz[0] - p.xyz[0],
                      xyz[1] - p.xyz[1],
                      xyz[2] - p.xyz[2]);
      }
      static float distsq(const pos_t &p1, const pos_t &p2) {
         float d0 = p2.xyz[0] - p1.xyz[0];
         float d1 = p2.xyz[1] - p1.xyz[1];
         float d2 = p2.xyz[2] - p1.xyz[2];
         float sum = d0*d0 + d1*d1 + d2*d2;
         return sum;
      }
   };

   std::vector<pos_t> positions;
#ifndef STANDALONE_INPUTS
   std::vector<mmdb::Atom *> atoms;
   atoms_container_t(mmdb::Manager *mol);
#endif // STANDALONE_INPUTS   

public:
   atoms_container_t() {}
   atoms_container_t(const std::string &file_name);
   void write_atoms(const std::string &file_name) const;
   unsigned int size() const { return positions.size(); }
   const pos_t &operator[](unsigned int idx) const { return positions[idx]; }
#ifndef STANDALONE_INPUTS
   void transfer_to(atom_selection_container_t &asc);
   coot::atom_spec_t spec(const int i) const { return coot::atom_spec_t(atoms[i]); }
#endif // STANDALONE_INPUTS

};

