
#include "CudaLBFGS/lbfgs.h"
#include "CudaLBFGS/error_checking.h"
#include "CudaLBFGS/timer.h"

#include <iostream>
#include <vector>
#include <string>

#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_runtime_api.h>
#include <device_launch_parameters.h>
#include <device_functions.h>

#include <clipper/core/coords.h>


#define cudaCheckErrors(msg) \
    do { \
        cudaError_t __err = cudaGetLastError(); \
        if (__err != cudaSuccess) { \
           fprintf(stderr, "Fatal error: %s (%s at %s:%d)\n",   \
                msg, cudaGetErrorString(__err), \
                __FILE__, __LINE__); \
           fprintf(stderr, "*** FAILED ***\n"); \
        } \
    } while (0)


#include <clipper/core/coords.h>

#define HAVE_GSL  // This is a hack to compile simple-restraint.hh
                  // we need a better mechanism
#define HAVE_CXX_THREAD // hack, why is this needed? doesn't it get set with -std=c++?
#include <coot/ideal/simple-restraint.hh>

#include "cuda_restraint.cuh"
#include "gpu-bonds.hh"


int main(int argc, char **argv) {

   int status = 1;

   std::vector<std::pair<bool,mmdb::Residue *> > residues;
   std::vector<mmdb::Link> links;
   coot::protein_geometry geom;
   mmdb::Manager *mol = 0;
   std::vector<coot::atom_spec_t> fixed_atom_specs;
   clipper::Xmap<float> xmap; // don't go out of scope until refinement is over

   coot::restraints_container_t restraints(residues, links, geom, mol, fixed_atom_specs, &xmap);

  coot::restraint_usage_Flags flags = coot::ALL_RESTRAINTS;
  int imol = 0;
  bool do_torsions = true;
  double rama_plot_weight = 1.0;
  bool do_rama_plot_restraints = true;
  coot::pseudo_restraint_bond_type sec_struct_pseudo_bonds = coot::NO_PSEUDO_BONDS;
  if (false) {
     unsigned int n_restraints = restraints.make_restraints(imol, geom, flags, do_torsions, true,
                                                            rama_plot_weight, do_rama_plot_restraints,
                                                            true, true, sec_struct_pseudo_bonds);
  }



   // atoms
   std::vector<clipper::Coord_orth> atom_positions;
   atom_positions.push_back(clipper::Coord_orth(1.5, 2.6, 3.1));
   atom_positions.push_back(clipper::Coord_orth(5.2, 6.3, 7.2));

   unsigned int n_atoms = atom_positions.size();

   gpu_bonds gb(n_atoms, restraints);
   lbfgs minimizer(gb);

   float *d_x; // the coordinates go here

   // allocate host memory:
   float *x = new float[n_atoms * 3];
   // allocate d_x on the device
   cudaMalloc(&d_x, n_atoms * 3 * sizeof(float));
   // now fill x
   for (unsigned int i=0; i<atom_positions.size(); i++) {
      x[3*i  ] = atom_positions[i].x();
      x[3*i+1] = atom_positions[i].y();
      x[3*i+2] = atom_positions[i].z();
      std::cout << "Input i: " << i << " x: " << x[3*i] << " " << x[3*i+1] << " " << x[3*i+2] << std::endl;
   }
   // copy x to d_x
   cudaMemcpy(d_x, x, n_atoms * 3 * sizeof(float), cudaMemcpyHostToDevice);

   std::cout << "Minimize..." << std::endl;
   minimizer.minimize(d_x);
   cudaMemcpy(x, d_x, n_atoms * 3  * sizeof(float), cudaMemcpyDeviceToHost);
   for (unsigned int i=0; i<atom_positions.size(); i++) {
      std::cout << " Result " << x[3*i] << " " << x[3*i+1] << " " << x[3*i+2] << std::endl;
   }
   float bl_x_2 = x[3] - x[0];
   float bl_y_2 = x[4] - x[1];
   float bl_z_2 = x[5] - x[2];
   float bl_2 = bl_x_2 * bl_x_2 + bl_y_2 * bl_y_2 + bl_z_2 * bl_z_2;
   float bl = sqrt(bl_2);
   std::cout << " Result bl " << bl << std::endl;

    cudaFree(d_x);
   cudaCheckErrors("cudaFree fail (d_x)");

   delete [] x;
   std::cout << "delete [] x done " << std::endl;

   std::cout << "returning to OS " << std::endl;
   return status;
}
