// -*-c++-*-
  #ifndef CLIPPER_CUDA_HH
  #define CLIPPER_CUDA_HH
 
  #include <clipper/core/coords.h>
  #include <clipper/core/nxmap.h>
 
  #include "Matrix3x3.cuh"
 
  // Note to self vec3_cuda should be templatized for int also.
  // then grid_cuda can inherit from the int verison of vec3_cuda
 
  class grid_cuda {
  public:
     __host__ __device__ grid_cuda() {}
     __host__ __device__ grid_cuda(int uu, int vv, int ww) : nu(uu), nv(vv), nw(ww) {}
     int nu;
     int nv;
     int nw;
 
  };
 
  class coord_grid_cuda {
  public:
     __host__ __device__ coord_grid_cuda() {}
     __host__ __device__ coord_grid_cuda(int uu, int vv, int ww) : u(uu), v(vv), w(ww) {}
     int u;
     int v;
     int w;
 
     __device__ coord_grid_cuda operator+(const coord_grid_cuda &trn) const {
        return coord_grid_cuda(u+trn.u, v+trn.v, w+trn.w);
     }
     __device__ coord_grid_cuda operator-(const coord_grid_cuda &trn) const {
        return coord_grid_cuda(u-trn.u, v-trn.v, w-trn.w);
     }
     __device__ inline coord_grid_cuda &next_u() { u++; return *this;}
     __device__ inline coord_grid_cuda &prev_u() { u--; return *this;}
     __device__ inline coord_grid_cuda &next_v() { v++; return *this;}
     __device__ inline coord_grid_cuda &prev_v() { v--; return *this;}
     __device__ inline coord_grid_cuda &next_w() { w++; return *this;}
     __device__ inline coord_grid_cuda &prev_w() { w--; return *this;}
 
  };
 
  class vec3_cuda {

     public:
     __host__ __device__ vec3_cuda() { vec[0] = 0; vec[1]=0; vec[2]=0; }
     __host__ __device__ vec3_cuda(float xx, float yy, float zz) { vec[0]=xx; vec[1]=yy; vec[2]=zz; }
     __host__ __device__ vec3_cuda(const vec3_cuda &v) { vec[0]=v.x(); vec[1]=v.y(); vec[2]=v.z(); }
     float vec[3];
     __host__ __device__ float x() const { return vec[0];}
     __host__ __device__ float y() const { return vec[1];}
     __host__ __device__ float z() const { return vec[2];}
     __device__ vec3_cuda operator+(const vec3_cuda &trn) const {
        return vec3_cuda(x()+trn.x(), y()+trn.y(), z()+trn.z());
     }
     __device__ vec3_cuda operator-(const vec3_cuda &trn) const {
        return vec3_cuda(x()-trn.x(), y()-trn.y(), z()-trn.z());
     }
     __device__ void operator+=(const vec3_cuda &trn) {
        vec[0] += trn.x();
        vec[1] += trn.y();
        vec[2] += trn.z();
     }
     __device__ void operator-=(const vec3_cuda &trn) {
        vec[0] -= trn.x();
        vec[1] -= trn.y();
        vec[2] -= trn.z();
     }
     __device__ vec3_cuda operator*(const float &f) const {
        return vec3_cuda(f*x(), f*y(), f*z());
     }
     __host__ __device__ float & operator[](int i) { return vec[i]; }
     __host__ __device__ float operator[](int i) const { return vec[i]; }
     __host__ __device__ float lengthsq() const { return vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2]; }
     __host__ __device__ static float dot(const vec3_cuda &o, const vec3_cuda &f) { return o.x()*f.x() + f.y()*o.y() + o.z()*f.z(); }
  };
 
  class coord_map_cuda : public vec3_cuda {
  public:
     __device__ __host__ coord_map_cuda(float uu, float vv, float ww) : vec3_cuda(uu, vv, ww) {}
     __host__ __device__ coord_map_cuda(const vec3_cuda &v) : vec3_cuda(v) {}
 
     // This is not as elegant as clipper!
     __device__ float u() const { return vec[0]; }
     __device__ float v() const { return vec[1]; }
     __device__ float w() const { return vec[2]; }
 
     // check floor(-1.3), floor(-1.9) - they need to round down. FIXME
     __host__ __device__ coord_map_cuda floor() const {
        return coord_map_cuda(::floor(static_cast<int>(vec[0])), ::floor(static_cast<int>(vec[1])), ::floor(static_cast<int>(vec[2])));
     }
  };
 
  class coord_frac_cuda : public vec3_cuda {
  public:
     __host__ __device__ coord_frac_cuda() : vec3_cuda() {};
     __host__ __device__ coord_frac_cuda(const vec3_cuda &v) : vec3_cuda(v) {};
     __host__ __device__ coord_frac_cuda(float xx, float yy, float zz) : vec3_cuda(xx,yy,zz) {}
  };
 
  class coord_orth_cuda : public vec3_cuda {
  public:
     __host__ __device__ coord_orth_cuda() : vec3_cuda() {};
     __host__ __device__ coord_orth_cuda(const vec3_cuda &v) : vec3_cuda(v) {};
     __host__ __device__ coord_orth_cuda(float xx, float yy, float zz) : vec3_cuda(xx,yy,zz) {}
  };
 
  class grad_map_cuda : public vec3_cuda {
  public:
     __host__ __device__ grad_map_cuda() : vec3_cuda() {};
     __host__ __device__ grad_map_cuda(const vec3_cuda &v) : vec3_cuda(v) {};
     __host__ __device__ grad_map_cuda(float xx, float yy, float zz) : vec3_cuda(xx,yy,zz) {}
     __device__ float du() const { return vec[0]; }
     __device__ float dv() const { return vec[1]; }
     __device__ float dw() const { return vec[2]; }
  };
 
  class grad_orth_cuda : public vec3_cuda {
  public:
     __host__ __device__ grad_orth_cuda() : vec3_cuda() {};
     __host__ __device__ grad_orth_cuda(const vec3_cuda &v) : vec3_cuda(v) {};
     __host__ __device__ grad_orth_cuda(float xx, float yy, float zz) : vec3_cuda(xx,yy,zz) {}
     __device__ float dx() const { return vec[0]; }
     __device__ float dy() const { return vec[1]; }
     __device__ float dz() const { return vec[2]; }
  };
 
 
  class mat33_cuda {
  public:
     float m[3][3]; // public for printing, can/should make private later
     __host__ __device__ mat33_cuda() {
 
       m[0][0] = 1.0;
       m[0][1] = 0.0;
       m[0][2] = 0.0;
       m[1][0] = 0.0;
       m[1][1] = 1.0;
       m[1][2] = 0.0;
       m[2][0] = 0.0;
       m[2][1] = 0.0;
       m[2][2] = 1.0;
 
     }
     __host__ __device__ mat33_cuda(const float &mm00, const float &mm01, const float &mm02,
                                    const float &mm10, const float &mm11, const float &mm12,
                                    const float &mm20, const float &mm21, const float &mm22) {
       m[0][0] = mm00;
       m[0][1] = mm01;
       m[0][2] = mm02;
       m[1][0] = mm10;
       m[1][1] = mm11;
       m[1][2] = mm12;
       m[2][0] = mm20;
       m[2][1] = mm21;
       m[2][2] = mm22;
      }
 
      __host__ __device__ mat33_cuda(const Matrix3x3_d &m33) {
            for (unsigned int i=0; i<3; i++) {
                  m[i][0] = m33.m_row[i].x;
                  m[i][1] = m33.m_row[i].y;
                  m[i][2] = m33.m_row[i].z;
            }
      }
      __device__ coord_orth_cuda operator*(const vec3_cuda &co) const {
 
           float x = co.x() * m[0][0] + co.y() * m[0][1] + co.z() * m[0][2];
           float y = co.x() * m[1][0] + co.y() * m[1][1] + co.z() * m[1][2];
           float z = co.x() * m[2][0] + co.y() * m[2][1] + co.z() * m[2][2];
           return coord_orth_cuda(x,y,z);
      }
 
      __device__ mat33_cuda operator*(const mat33_cuda &m2) const {
         return mat33_cuda ( m[0][0]*m2.m[0][0] + m[0][1]*m2.m[1][0] + m[0][2]*m2.m[2][0],
                             m[0][0]*m2.m[0][1] + m[0][1]*m2.m[1][1] + m[0][2]*m2.m[2][1],
                             m[0][0]*m2.m[0][2] + m[0][1]*m2.m[1][2] + m[0][2]*m2.m[2][2],
                             m[1][0]*m2.m[0][0] + m[1][1]*m2.m[1][0] + m[1][2]*m2.m[2][0],
                             m[1][0]*m2.m[0][1] + m[1][1]*m2.m[1][1] + m[1][2]*m2.m[2][1],
                             m[1][0]*m2.m[0][2] + m[1][1]*m2.m[1][2] + m[1][2]*m2.m[2][2],
                             m[2][0]*m2.m[0][0] + m[2][1]*m2.m[1][0] + m[2][2]*m2.m[2][0],
                             m[2][0]*m2.m[0][1] + m[2][1]*m2.m[1][1] + m[2][2]*m2.m[2][1],
                             m[2][0]*m2.m[0][2] + m[2][1]*m2.m[1][2] + m[2][2]*m2.m[2][2]);
      }
 
     // use this after a default constructor
     __device__ void zero() {
 
       m[0][0] = 0.0;
       m[1][1] = 0.0;
       m[2][2] = 0.0;
     }

     float get(const unsigned int &i, const unsigned &j) const { return m[i][j]; }
 
     // angles in radians, of course, zyz mode
     __device__ void rotate_euler(float r1, float r2, float r3) {
        rotate_z(r3);
        rotate_y(r2);
        rotate_z(r1);
     }
     __device__ void rotate_x(float f) {
        float mm[3][3];
        float mt[3][3];
        float cos_alpha = cosf(f);
        float sin_alpha = sinf(f);
        mm[0][0] = 1.0;
        mm[0][1] = 0.0;
        mm[0][2] = 0.0;
        mm[1][0] = 0.0;
        mm[1][1] = cos_alpha;
        mm[1][2] = -sin_alpha;
        mm[2][0] = 0.0;
        mm[2][1] = sin_alpha;
        mm[2][2] = cos_alpha;
        for (int i=0; i<3; i++) {
           for (int j=0; j<3; j++) {
              float sum = 0.0;
              for (int k=0; k<3; k++) {
                 sum += m[k][j] * mm[k][i];
              }
              mt[i][j] = sum;
           }
        }
        for (int i=0; i<3; i++)
           for (int j=0; j<3; j++)
              m[i][j] = mt[i][j];
     }
     __device__ void rotate_y(float f) {
        float mm[3][3];
        float mt[3][3];
        float cos_alpha = cosf(f);
        float sin_alpha = sinf(f);
        mm[0][0] = cos_alpha;
        mm[0][1] = 0.0;
        mm[0][2] = -sin_alpha;
        mm[1][0] = 0.0;
        mm[1][1] = 1.0;
        mm[1][2] = 0.0;
        mm[2][0] = sin_alpha;
        mm[2][1] = 0.0;
        mm[2][2] = cos_alpha;
        for (int i=0; i<3; i++) {
           for (int j=0; j<3; j++) {
              float sum = 0.0;
              for (int k=0; k<3; k++) {
                 sum += m[k][j] * mm[k][i];
              }
              mt[i][j] = sum;
           }
        }
        for (int i=0; i<3; i++)
           for (int j=0; j<3; j++)
              m[i][j] = mt[i][j];
     }
     __device__ void rotate_z(float f) {
        float mm[3][3];
        float mt[3][3];
        float cos_alpha = cosf(f);
        float sin_alpha = sinf(f);
        mm[0][0] = cos_alpha;
        mm[0][1] = -sin_alpha;
        mm[0][2] = 0.0;
        mm[1][0] = sin_alpha;
        mm[1][1] = cos_alpha;
        mm[1][2] = 0.0;
        mm[2][0] = 0.0;
        mm[2][1] = 0.0;
        mm[2][2] = 1.0;
        for (int i=0; i<3; i++) {
           for (int j=0; j<3; j++) {
              float sum = 0.0;
              for (int k=0; k<3; k++) {
                 sum += m[k][j] * mm[k][i];
              }
              mt[i][j] = sum;
           }
        }
        for (int i=0; i<3; i++)
           for (int j=0; j<3; j++)
              m[i][j] = mt[i][j];
     }
 
    //  __device__ float3 eigens(bool sort);
  #include "clipper-cuda-eigens.cuh"
 
     __host__ __device__ void print_self() const {
        printf("| %6.3f %6.3f %6.3f |\n", m[0][0], m[0][1], m[0][2]);
        printf("| %6.3f %6.3f %6.3f |\n", m[1][0], m[1][1], m[1][2]);
        printf("| %6.3f %6.3f %6.3f |\n", m[2][0], m[2][1], m[2][2]);
     }
 
  };
 
  class rt_cuda {
     public:
     __host__ __device__ rt_cuda() { rot = mat33_cuda(); trans = coord_orth_cuda(); }
     __host__ __device__ rt_cuda(const mat33_cuda &m, const coord_orth_cuda &co) : rot(m), trans(co) {}
     mat33_cuda rot;
     coord_orth_cuda trans;
     __device__ vec3_cuda operator* (const coord_orth_cuda &co) const {
        return rot * co + trans;
     }
     __host__ __device__ void print_self() const {
       printf("| %6.3f %6.3f %6.3f |\n", rot.m[0][0], rot.m[0][1], rot.m[0][2]);
       printf("| %6.3f %6.3f %6.3f |\n", rot.m[1][0], rot.m[1][1], rot.m[1][2]);
       printf("| %6.3f %6.3f %6.3f |\n", rot.m[2][0], rot.m[2][1], rot.m[2][2]);
       printf("( %8.3f %8.3f %8.3f )\n", trans.x(), trans.y(), trans.z());
    }
    __host__ float get_rot(const unsigned int &i, const unsigned int &j) const { return rot.get(i,j); }
    __host__ float get_trn(int i) const { return trans[i]; }
  };
 
  class orth_grid_4x4_cuda {
     float m[4][4];
  public:
     orth_grid_4x4_cuda(const float &mm00, const float &mm01, const float &mm02, const float &mm03, 
                        const float &mm10, const float &mm11, const float &mm12, const float &mm13, 
                        const float &mm20, const float &mm21, const float &mm22, const float &mm23, 
                        const float &mm30, const float &mm31, const float &mm32, const float &mm33) {
       m[0][0] = mm00;
       m[0][1] = mm01;
       m[0][2] = mm02;
       m[0][3] = mm03;
       m[1][0] = mm10;
       m[1][1] = mm11;
       m[1][2] = mm12;
       m[1][3] = mm13;
       m[2][0] = mm20;
       m[2][1] = mm21;
       m[2][2] = mm22;
       m[2][3] = mm23;
       m[3][0] = mm30;
       m[3][1] = mm31;
       m[3][2] = mm32;
       m[3][3] = mm33;
      }
  };
 
class simple_atom_cuda {
public:
   __host__ __device__ simple_atom_cuda(const coord_orth_cuda &c, float w) : coord(c), weight(w) {}
   coord_orth_cuda coord;
   float weight;
};


#endif // CLIPPER_CUDA_HH
