
// -*-c++-*-

#include <iostream>
#include <vector>
#include <string>

#include <clipper/core/coords.h>
#include <clipper/ccp4/ccp4_map_io.h>

#include <coot/utils/coot-utils.hh>
#include <coot/utils/split-indices.hh>
#include <coot/coot-utils/coot-map-utils.hh>
#include <coot/ideal/simple-restraint.hh>

#include "coot-cuda-minimize.hh"

         void contacts_for_range(mmdb::Contact *pscontact,
                                 const std::pair<unsigned int, unsigned int> &contact_range,
                                 std::vector<coot::extra_restraints_t::extra_bond_restraint_t> *ebr,
                                 mmdb::PPAtom SelAtom,
                                 int UDDAtomIndexHandle,
                                 const std::map<std::string, std::vector<std::pair<std::string, std::string> > > &bonded_neighbours);


void contacts_for_range(mmdb::Contact *pscontact,
                        const std::pair<unsigned int, unsigned int> &contact_range,
                        std::vector<coot::extra_restraints_t::extra_bond_restraint_t> *ebr,
                        mmdb::PPAtom SelAtom,
                        int UDDAtomIndexHandle,
                        const std::map<std::string, std::vector<std::pair<std::string, std::string> > > &bonded_neighbours) {

   // std::cout << "------ contacts_for_range() " << bonded_neighbours.size() << std::endl;

   for (unsigned int i=contact_range.first; i<contact_range.second; i++) {
      mmdb::Atom *at_1 = SelAtom[pscontact[i].id1];
      mmdb::Atom *at_2 = SelAtom[pscontact[i].id2];
      std::string ele_1 = at_1->element;
      std::string ele_2 = at_2->element;
      if (ele_1 != " H" && ele_2 != " H") {
	 bool ignore_this = false; // set for bonded and angled atoms
	 bool in_same_res = false;
	 if (at_1->residue == at_2->residue)
	    in_same_res = true;

	 if (in_same_res) {
	    std::string comp_id = at_1->residue->GetResName();
	    std::string at_name_1 = at_1->GetAtomName();
	    std::string at_name_2 = at_2->GetAtomName();

	    // This is slow
	    // if (geom.are_bonded_or_angle_related(comp_id, at_name_1, at_name_2))
	    // ignore_this = true;
	    //

	    std::map<std::string, std::vector<std::pair<std::string, std::string> > > ::const_iterator it = bonded_neighbours.find(comp_id);

	    if (it != bonded_neighbours.end()) {
	       const std::vector<std::pair<std::string, std::string> > &bps = it->second;

	       for (unsigned int ipr=0; ipr<bps.size(); ipr++) {
		  if (at_name_1 == bps[ipr].first) {
		     if (at_name_2 == bps[ipr].second) {
			ignore_this = true;
			break;
		     }
		  }
		  if (at_name_2 == bps[ipr].first) {
		     if (at_name_1 == bps[ipr].second) {
			ignore_this = true;
			break;
		     }
		  }
	       }
	    } else {
	       std::cout << "error:: failed to find " << comp_id << " in bonded_neighbours" << std::endl;
	    }
	 } else {
	    std::string at_name_1 = at_1->GetAtomName();
	    std::string at_name_2 = at_2->GetAtomName();
	    // by hand (hmmm)
	    // PDBv3 FIXME
	    if ((at_name_1 == " N  " && at_name_2 == " C  ") ||
		(at_name_1 == " C  " && at_name_2 == " N  ") ||
		(at_name_1 == " N  " && at_name_2 == " O  ") ||
		(at_name_1 == " N  " && at_name_2 == " CA ") ||
		(at_name_1 == " O  " && at_name_2 == " N  ") ||
		(at_name_1 == " CA " && at_name_2 == " N  ")) {
	       ignore_this = true;
	    }
	 }

	 if (false) // are you using the correct pdb file? :-)
	    std::cout << "Here with ignore this " << ignore_this << " "
		      << coot::atom_spec_t(at_1) << " "
		      << coot::atom_spec_t(at_2) << std::endl;

	 if (! ignore_this) {
	    clipper::Coord_orth p1 = coot::co(at_1);
	    clipper::Coord_orth p2 = coot::co(at_2);
	    double dist = sqrt((p1-p2).lengthsq());
	    double esd  = 0.05;
	    coot::atom_spec_t atom_spec_1(at_1);
	    coot::atom_spec_t atom_spec_2(at_2);
	    int idx_1 = -1;
	    int idx_2 = -1;
	    at_1->GetUDData(UDDAtomIndexHandle, idx_1);
	    at_2->GetUDData(UDDAtomIndexHandle, idx_2);
	    atom_spec_1.int_user_data = idx_1;
	    atom_spec_2.int_user_data = idx_2;
	    coot::extra_restraints_t::extra_bond_restraint_t br(atom_spec_1, atom_spec_2, dist, esd);
	    ebr->push_back(br);
	 }
      }
   }
}


void
generate_local_self_restraints(atom_selection_container_t asc,
                               int selHnd, float local_dist_max,
			       const coot::protein_geometry &geom,
			       coot::extra_restraints_t *extra_restraints_p) {

   coot::extra_restraints_t &extra_restraints = *extra_restraints_p;

   // clear what's already there - if anything
   // extra_restraints.bond_restraints.clear(); - no.

   // extra_restraints.bond_restraints.reserve(asc.n_selected_atoms * 44);

   int nSelAtoms;
   mmdb::PPAtom SelAtom;
   asc.mol->GetSelIndex(selHnd, SelAtom, nSelAtoms);

   std::cout << "DEBUG:: generate_local_self_restraints() here with nSelAtoms " << nSelAtoms << std::endl;

   // bonded_neighbours in this case, means bonded or angle-related
   // bonded_neighbours["ALA"] -> all bond pairs and 1-3 angles
   std::map<std::string, std::vector<std::pair<std::string, std::string> > > bonded_neighbours;

   // now find contacts:
   //
   mmdb::Contact *pscontact = NULL;
   int n_contacts;
   long i_contact_group = 1;
   mmdb::mat44 my_matt;
   for (int i=0; i<4; i++)
      for (int j=0; j<4; j++)
         my_matt[i][j] = 0.0;
   for (int i=0; i<4; i++) my_matt[i][i] = 1.0;
   //
   std::cout << "debug:: seeking contacts " << nSelAtoms << " atoms\n";
   asc.mol->SeekContacts(SelAtom,   nSelAtoms,
			      SelAtom,   nSelAtoms,
			      0.1, local_dist_max, // min, max distances
			      0,        // seqDist 0 -> in same res also
			      pscontact, n_contacts,
			      0, &my_matt, i_contact_group);

   // SeekContacts takes a second or so for beasty.

   std::cout << "debug:: processing contacts " << n_contacts << " contacts\n";
   if (n_contacts > 0) {
      if (pscontact) {

         // Divide n_contacts into ranges and use threads. But then how do we consolidate?
         // Hmm. Perhaps call restraints.add_extra_restraints(imol, extras, geom)
         // multiple times (extras -> extras_vec[i_thread]). That might be slow.

         unsigned int n_threads = coot::get_max_number_of_threads()-1;
         if (n_threads == 0) n_threads = 1;
         ctpl::thread_pool thread_pool(n_threads);
         std::vector<std::pair<unsigned int, unsigned int> > contacts_ranges =
         coot::atom_index_ranges(n_contacts, n_threads);

         // put results here, split by thread
         std::vector<std::vector<coot::extra_restraints_t::extra_bond_restraint_t> > ebrv(n_threads);

         // fill bonded neighbours

         std::map<std::string, std::vector<std::pair<std::string, std::string> > >::const_iterator it;

         std::vector<std::string> comp_ids = geom.monomer_restraints_comp_ids();
         for (unsigned int ii=0; ii<comp_ids.size(); ii++) {
            const std::string &comp_id = comp_ids[ii];
            int imol_no = -1;
            std::vector<std::pair<std::string, std::string> > bps = geom.get_bonded_and_1_3_angles(comp_id, imol_no);
            bonded_neighbours[comp_id] = bps;
         }

         std::cout << "Generating local distance restraints with " << n_threads << " threads " << std::endl;
         std::vector<std::thread> threads;
         for (unsigned int ith=0; ith<n_threads; ith++) {
            // contacts_for_range(pscontact, std::cref(contacts_ranges[ith]), &ebrv[ith], SelAtom, asc.UDDAtomIndexHandle, bonded_neighbours);
            threads.push_back(std::thread(contacts_for_range, pscontact, std::cref(contacts_ranges[ith]), &ebrv[ith], SelAtom, asc.UDDAtomIndexHandle, std::cref(bonded_neighbours)));
         }
         for (unsigned int ith=0; ith<n_threads; ith++) {
            threads[ith].join();
         }

         std::cout << "Inserting extra bond restraints" << std::endl;
         for (std::size_t i=0; i<n_threads; i++)
            std::move(ebrv[i].begin(), ebrv[i].end(), std::back_inserter(extra_restraints.bond_restraints));
         std::cout << "Done inserting extra bond restraints" << std::endl;

      } // pscontact
   } // n_contacts
   std::cout << "DEBUG:: generate_local_self_restraints() Made " << extra_restraints.bond_restraints.size()
             << " extra bond restraints" << std::endl;

}

std::vector<std::string>
get_chain_ids(mmdb::Manager *mol) {

   std::vector<std::string> v;
   int imod=1;
   mmdb::Model *model_p = mol->GetModel(imod);
   if (model_p) {
      int n_chains = model_p->GetNumberOfChains();
      for (int ichain=0; ichain<n_chains; ichain++) {
         mmdb::Chain *chain_p = model_p->GetChain(ichain);
         if (chain_p) {
             v.push_back(chain_p->GetChainID());
         }
      }
   }
   return v;
}


void
generate_local_self_restraints(atom_selection_container_t asc,
                               float local_dist_max,
			       const std::string &chain_id,
			       const coot::protein_geometry &geom,
			       coot::extra_restraints_t *extra_restraints_p) {

   // Find all the contacts in chain_id that are less than or equal to local_dist_max
   // that are not bonded or related by an angle.

   int selHnd = asc.mol->NewSelection(); // - check the deletion

   asc.mol->SelectAtoms(selHnd, 0, chain_id.c_str(),
			     mmdb::ANY_RES, "*", // start, insertion code
			     mmdb::ANY_RES, "*", // end, insertion code
			     "*", // residue name
			     "*",
			     "*", // elements
			     "*"); // alt locs

   generate_local_self_restraints(asc, selHnd, local_dist_max, geom, extra_restraints_p);

   asc.mol->DeleteSelection(selHnd);
}

void refine(const std::string &pdb_file_name,
            const std::string &mtz_file_name, const std::string &fwt, const std::string &phwt,
            const std::string &map_file_name,
            float map_weight,
            float geman_mcclure_cutoff,
            float geman_mcclure_alpha,
            int n_blocks, int n_thread_per_block,
            const std::string &pdbout_file_name,
            const std::string &cifout_file_name,
            bool write_data_for_standalone,
	    bool progressive_unblur_mode,
	    int progressive_unblur_n_cycles,
            float progressive_unblur_max_blur) {

   if (true) {

      bool map_is_good = false; // currently

      bool allow_duplseqnum = true;
      bool convert_to_v2_name_flag = false;
      atom_selection_container_t asc = get_atom_selection(pdb_file_name,
							  allow_duplseqnum,
							  convert_to_v2_name_flag);
      std::cout << "DEBUG:: here with n links " << asc.links.size() << std::endl;

      if (false) { // debugging the names of the input molecule (missing restraints)
            for(int imod = 1; imod<=asc.mol->GetNumberOfModels(); imod++) {
	            mmdb::Model *model_p = asc.mol->GetModel(imod);
	    if (model_p) {
	       int n_chains = model_p->GetNumberOfChains();
	       for (int ichain=0; ichain<n_chains; ichain++) {
		  mmdb::Chain *chain_p = model_p->GetChain(ichain);
		  std::cout << "  chain  " << chain_p->GetChainID() << std::endl;
		  int nres = chain_p->GetNumberOfResidues();
		  for (int ires=0; ires<nres; ires++) {
		     mmdb::Residue *residue_p = chain_p->GetResidue(ires);
		     std::cout << "    " << coot::residue_spec_t(residue_p) << " "
		               << residue_p->GetResName() << std::endl;
		     int n_atoms = residue_p->GetNumberOfAtoms();
		     for (int iat=0; iat<n_atoms; iat++) {
		        mmdb::Atom *at = residue_p->GetAtom(iat);
		        std::cout << "        " << coot::atom_spec_t(at) << std::endl;
		     }
		  }
	       }
	    }
	 }
      }

      if (asc.read_success) {
	 asc.add_old_atom_indices();
	 std::vector<std::pair<bool,mmdb::Residue *> > residues;
	 int imod = 1;
	 mmdb::Model *model_p = asc.mol->GetModel(imod);
	 if (model_p) {
	    int n_chains = model_p->GetNumberOfChains();
	    for (int ichain=0; ichain<n_chains; ichain++) {
	       mmdb::Chain *chain_p = model_p->GetChain(ichain);
	       int nres = chain_p->GetNumberOfResidues();
	       for (int ires=0; ires<nres; ires++) {
	          mmdb::Residue *residue_p = chain_p->GetResidue(ires);
	          if (residue_p) {
	             std::pair<bool,mmdb::Residue *> p(false, residue_p);
	             residues.push_back(p);
	          }
	       }
	    }
         }

         clipper::Xmap<float> xmap;

         if (map_file_name.length() > 0) {
	    if (coot::file_exists(map_file_name)) {
	       // should we try/catch around this read?
	       clipper::CCP4MAPfile file;
	       file.open_read(map_file_name);
	       file.import_xmap(xmap);
	       file.close_read();
	       map_is_good = true;
	    } else {
	       std::cout << "WARNING:: map file " << map_file_name << " not found"
	                 << std::endl;
	    }
         } else {
           if (!mtz_file_name.empty() && !fwt.empty() && !phwt.empty()) {
              coot::util::map_fill_from_mtz(&xmap, mtz_file_name, fwt, phwt, "", 0, 0);
           }
         }

         if (xmap.is_null()) return;

         coot::protein_geometry geom;
         geom.set_verbose(false);
         geom.init_standard();
         geom.add_planar_peptide_restraint();
         mmdb::Manager *mol = asc.mol;
         int read_number = 55;
         int imol = 0;
         geom.read_extra_dictionaries_for_molecule(mol, imol, &read_number);

         coot::restraints_container_t restraints(residues, geom, mol, &xmap);

         unsigned int n_threads = coot::get_max_number_of_threads();
         ctpl::thread_pool thread_pool(n_threads);
         if (n_threads > 0) {
            restraints.thread_pool(&thread_pool, n_threads);
         } else {
            std::cout << "No threads - Stop now" << std::endl;
            return;
         }

         coot::restraint_usage_Flags flags = coot::ALL_RESTRAINTS;
         flags = coot::TYPICAL_RESTRAINTS; // torsion restraints are slow to configure and don't work yet.
         bool do_torsions = true;
         double rama_plot_weight = 1.0;
         bool do_rama_plot_restraints = false;
         bool do_GM_restraints_by_chain = false;
         bool do_GM_restraints = true;
         coot::pseudo_restraint_bond_type sec_struct_pseudo_bonds = coot::NO_PSEUDO_BONDS;

         std::cout << "DEBUG:: in refine() make_restraints() " << std::endl;
         unsigned int nn = restraints.df_by_thread_results_size();
         unsigned int n_restraints = restraints.make_restraints(imol, geom, flags,
								do_torsions, true, rama_plot_weight,
								do_rama_plot_restraints,
								true, true, sec_struct_pseudo_bonds);

         std::cout << "DEBUG:: in refine() n_restraints: " << n_restraints << std::endl;

         if (do_GM_restraints) {
             if (do_GM_restraints_by_chain) {
                std::vector<std::string> chain_ids = get_chain_ids(asc.mol);
                for (auto chid : chain_ids) {
                   coot::extra_restraints_t extras;
                   std::cout << "Generating local self restraints for the " << chid << " chain" << std::endl;
                   generate_local_self_restraints(asc, geman_mcclure_cutoff, chid, geom, &extras); // fill extras
                   restraints.set_geman_mcclure_alpha(geman_mcclure_alpha);
                   restraints.add_extra_restraints(imol, extras, geom);
                }
            } else {
               coot::extra_restraints_t extras;
               int selHnd = asc.mol->NewSelection();
               asc.mol->SelectAtoms(selHnd, 0, "*",
               mmdb::ANY_RES, "*", // start, insertion code
               mmdb::ANY_RES, "*", // end, insertion code
               "*", // residue name
               "*",
               "*", // elements
               "*"); // alt locs
               auto tp_0 = std::chrono::high_resolution_clock::now();
               generate_local_self_restraints(asc, selHnd, geman_mcclure_cutoff, geom, &extras);
               auto tp_1 = std::chrono::high_resolution_clock::now();
               restraints.add_extra_restraints(imol, extras, geom);
               auto tp_2 = std::chrono::high_resolution_clock::now();
               auto d10 = std::chrono::duration_cast<std::chrono::milliseconds>(tp_1 - tp_0).count();
               auto d21 = std::chrono::duration_cast<std::chrono::milliseconds>(tp_2 - tp_1).count();
               std::cout << "---------------------- timings for extra restraints: generate: " << d10 << " add: " << d21
                         << " milliseconds" << std::endl;
               asc.mol->DeleteSelection(selHnd);
            }
         }

         // If I add these, then I should change the name of this file to .cu?
         // How about calling a function that does this - that function being in a .cu file?
         // gpu_bonds gb(n_atoms, restraints);
         // lbfgs minimizer(gb);
         //
         // cuda_minimize() writes files for the moment.


         if (! progressive_unblur_mode) {

            cuda_minimize(restraints, asc, xmap, map_weight, n_blocks, n_thread_per_block,
                          write_data_for_standalone);

         } else {

            std::pair<float, float> mv = coot::util::mean_and_variance(xmap);
            float map_sigma_orig = sqrt(mv.second);

            for (int i=progressive_unblur_n_cycles; i>=0; i--) {
               // when i =0; blur = 0
               // when i = progressive_blur_n_cycles, blur = progressive_blur_max_blur
               float b_factor = static_cast<float>(i) * progressive_unblur_max_blur/static_cast<float>(progressive_unblur_n_cycles);
               std::cout << "### Cycle i " << i << " b_factor " << b_factor << std::endl;
               clipper::Xmap<float> blurred_xmap = coot::util::sharpen_blur_map(xmap,b_factor);
               std::pair<float, float> mv = coot::util::mean_and_variance(xmap);
               float map_sigma_blurred = sqrt(mv.second);
               float map_weight_blurred = map_weight * map_sigma_blurred / map_sigma_orig;
               cuda_minimize(restraints, asc, blurred_xmap, map_weight_blurred, n_blocks, n_thread_per_block,
                             write_data_for_standalone);
            }
        }
        if (!pdbout_file_name.empty())
             asc.mol->WritePDBASCII(pdbout_file_name.c_str());
        if (!cifout_file_name.empty())
             asc.mol->WriteCIFASCII(cifout_file_name.c_str());
         std::cout << "---- Post refinement ---- " << std::endl;
         restraints.analyze_for_bad_restraints();
      }
   }
}

class new_variables_t {
   void parse_command_line(int argc, char **argv) {
      nb.first = false;
      nt.first = false;
      mw.first = false;
      geman_mcclure_alpha.first = false;
      geman_mcclure_cutoff.first = false;
      write_data.first = false;
      pdbout_file_name.first = false;
      cifout_file_name.first = false;
      for (unsigned int i=1; i<(argc-1); i++) {
         std::string key = argv[i];
         std::string val = argv[i+1];

         // std::cout << "debug key " << key << " value " << val << std::endl;

         if (key == "map")       map_file_name = std::pair<bool, std::string> (true, val);
         if (key == "mapin")     map_file_name = std::pair<bool, std::string> (true, val);
         if (key == "pdb")       pdb_file_name = std::pair<bool, std::string> (true, val);
         if (key == "pdbin")     pdb_file_name = std::pair<bool, std::string> (true, val);
         if (key == "cif")       pdb_file_name = std::pair<bool, std::string> (true, val);
         if (key == "cifin")     pdb_file_name = std::pair<bool, std::string> (true, val);
         if (key == "mtz")       mtz_file_name = std::pair<bool, std::string> (true, val);
         if (key == "mtzin")     mtz_file_name = std::pair<bool, std::string> (true, val);
         if (key == "pdbout") pdbout_file_name = std::pair<bool, std::string> (true, val);
         if (key == "cifout") cifout_file_name = std::pair<bool, std::string> (true, val);
         if (key == "f")                    f = std::pair<bool, std::string> (true, val);
         if (key == "phi")                phi = std::pair<bool, std::string> (true, val);
         if (key == "nthreads")   nt          = std::pair<bool, unsigned int> (true, coot::util::string_to_int(val));
         if (key == "nblocks")    nb          = std::pair<bool, unsigned int> (true, coot::util::string_to_int(val));
         if (key == "map-weight") mw          = std::pair<bool, float> (true, coot::util::string_to_float(val));
         if (key == "alpha")                geman_mcclure_alpha  = std::pair<bool, float> (true, coot::util::string_to_float(val));
         if (key == "geman-mcclure-alpha")  geman_mcclure_alpha  = std::pair<bool, float> (true, coot::util::string_to_float(val));
         if (key == "geman-mcclure-cutoff") geman_mcclure_cutoff = std::pair<bool, float> (true, coot::util::string_to_float(val));
         if (key == "write-data")      write_data = std::pair<bool, bool> (true, coot::util::string_to_int(val));
         if (key == "progressive-unblur-mode") progressive_blur_mode = std::pair<bool, bool> (true, true);
         if (key == "progressive-unblur-n-cycles")   progressive_blur_n_cycles = std::pair<bool, float> (true, coot::util::string_to_int(val));
         if (key == "progressive-unblur-max-blur")   progressive_blur_max_blur = std::pair<bool, float> (true, coot::util::string_to_float(val));
      }
   }
public:
   new_variables_t(int argc, char **argv) {
      parse_command_line(argc, argv);
   }
   std::pair<bool,bool> write_data;
   std::pair<bool,unsigned int> nb;
   std::pair<bool,unsigned int> nt;
   std::pair<bool,float> mw;
   std::pair<bool,float> geman_mcclure_alpha;
   std::pair<bool,float> geman_mcclure_cutoff;
   std::pair<bool,std::string> pdb_file_name;
   std::pair<bool,std::string> map_file_name;
   std::pair<bool,std::string> mtz_file_name;
   std::pair<bool,std::string> pdbout_file_name;
   std::pair<bool,std::string> cifout_file_name;
   std::pair<bool,std::string> f;
   std::pair<bool,std::string> phi;
   std::pair<bool,bool> progressive_blur_mode;
   std::pair<bool,int> progressive_blur_n_cycles;
   std::pair<bool,float> progressive_blur_max_blur;

};


int main(int argc, char **argv) {

   /*
   for (unsigned int ii=0; ii<128; ii++) {
      int r = ii % 4;
      std::cout << ii << " " << r << std::endl;
      }
   */

   int status = 0;
   std::string pdbout_file_name = "refined.pdb";
   std::string cifout_file_name;
   std::string pdb_file_name;
   std::string map_file_name;
   std::string mtz_file_name;
   bool progressive_blur_mode = false;
   int progressive_blur_n_cycles = 5;
   float progressive_blur_max_blur = 500.0;
   std::string fwt;
   std::string phwt;
   bool write_data_for_standalone = false;

   int nb = 1;
   int nt = 32;
   float map_weight = 10.0;
   float geman_mcclure_alpha = 0.01;
   float geman_mcclure_cutoff = 5.0;

   try {
      new_variables_t nv(argc, argv);
      if (nv.nb.first) nb         = nv.nb.second;
      if (nv.nt.first) nt         = nv.nt.second;
      if (nv.mw.first) map_weight = nv.mw.second;
      if (nv.geman_mcclure_alpha.first) geman_mcclure_alpha = nv.geman_mcclure_alpha.second;
      if (nv.geman_mcclure_cutoff.first) geman_mcclure_cutoff = nv.geman_mcclure_cutoff.second;
      if (nv.pdb_file_name.first) pdb_file_name = nv.pdb_file_name.second;
      if (nv.map_file_name.first) map_file_name = nv.map_file_name.second;
      if (nv.mtz_file_name.first) mtz_file_name = nv.mtz_file_name.second;
      if (nv.f.first)    fwt = nv.f.second;
      if (nv.phi.first) phwt = nv.phi.second;
      if (nv.write_data.first) write_data_for_standalone = true;
      if (nv.pdbout_file_name.first) pdbout_file_name = nv.pdbout_file_name.second;
      if (nv.cifout_file_name.first) cifout_file_name = nv.cifout_file_name.second;
      if (nv.progressive_blur_mode.first) progressive_blur_mode = true;
      if (nv.progressive_blur_n_cycles.first) progressive_blur_n_cycles = nv.progressive_blur_n_cycles.second;
      if (nv.progressive_blur_max_blur.first) progressive_blur_max_blur = nv.progressive_blur_max_blur.second;

      refine(pdb_file_name, mtz_file_name, fwt, phwt, map_file_name, map_weight, geman_mcclure_cutoff, geman_mcclure_alpha, nb, nt,
             pdbout_file_name, cifout_file_name, write_data_for_standalone,
             progressive_blur_mode, progressive_blur_n_cycles, progressive_blur_max_blur);
   }
   catch (const std::runtime_error &rte) {
      std::cout << "Error in input: " << rte.what() << std::endl;
   }

   return status;
}
