#include <iostream>
#include <iomanip>
#include <string>
#include <vector>
#include <chrono>

#include <math.h>
#include <stdio.h>
#include "cufft.h"

#include <clipper/core/coords.h>
#include <clipper/core/nxmap.h>
#include <clipper/core/xmap.h>
#include <clipper/ccp4/ccp4_map_io.h>

#define NX 192
#define NY 96
#define NZ 48

// #define NX 64
// #define NY 92
// #define NZ 150

#define BATCH 1 // was 10
#define NRANK 3

cufftHandle plan;
// cufftComplex *data;
int n[NRANK] = {NX, NY, NZ};


void write_map(cufftComplex *data_host, const clipper::Cell &cell, const clipper::Grid &grid, float sf, const std::string &map_file_name);

// return grid sampling and maybe other map details, like start and stop points.
class map_file_info_t {
public:
  map_file_info_t(const clipper::Grid &grid_in,
                  const clipper::Cell &cell_in,
                  cufftComplex *data_in) :
                  grid(grid_in), cell(cell_in), data(data_in) {}
  clipper::Grid grid;
  cufftComplex *data;
  clipper::Cell cell;
};

map_file_info_t
fill_data_from_map_file(const std::string &map_file_name);

void do_fft(cufftComplex *data_host, 
            const clipper::Cell &cell,
            const clipper::Grid &grid,
            const std::string &map_file_name);

void blue_pill();
void do_molecule_map(const std::string &map_file_name);

int main(int argc, char **argv) {

   // blue_pill();

   std::string map_file_name = "6e1h-fragment.map";
   if (argc > 1)
      map_file_name = argv[1];
   do_molecule_map(map_file_name);

   exit(0);
}


void blue_pill() {
   bool output_map_flag = true;

   cufftHandle plan;
   cufftComplex *data_dev;
   cufftComplex *data_host = static_cast<cufftComplex *> (new cufftComplex[NX*NY*NZ]);


   // Fill data here

   unsigned int idx_mid_x = NX/2;
   unsigned int idx_mid_y = NY/2;
   unsigned int idx_mid_z = NZ/2;
   for (unsigned int i=0; i<NX; i++) {
      float fx = (static_cast<float>(i)-static_cast<float>(idx_mid_x))/static_cast<float>(idx_mid_x); // -1 to +1 as float
      for (unsigned int j=0; j<NY; j++) {
         float fy = (static_cast<float>(j)-static_cast<float>(idx_mid_y))/static_cast<float>(idx_mid_y); // -1 to +1 as float
         for (unsigned int k=0; k<NZ; k++) {
            unsigned int idx = k + NZ * j + NZ * NY * i; // or the other way round?
            float fz = (static_cast<float>(k)-static_cast<float>(idx_mid_z))/static_cast<float>(idx_mid_z); // -1 to +1 as float
            float r2 = fx * fx + fy * fy + fz * fz;
            float x = 0.0;
            if (r2 < 0.35) {
               x = exp(-r2) * 10.0;
            }

            x = 100.0 * sin(fy * 6.28 * 2.01);

            data_host[idx].x = x;
            data_host[idx].y = 0.0;
         }
      }
   }

   bool output_debugging = false;
   if (output_debugging) {
      for (unsigned int i=0; i<NX; i++) {
         for (unsigned int j=0; j<NY; j++) {
            for (unsigned int k=0; k<NZ; k++) {
               unsigned int idx = k + NZ * j + NZ * NY * i; // or the other way round?
               float re = data_host[idx].x;
               float im = data_host[idx].y;
               float r2 = i*i + j*j + k+k;
               std::cout << "original-data " << i << " " << j << " " << k << " " << r2 << " " << re << " " << im << std::endl;
            }
         }
      }
   }


   clipper::Grid grid(NX, NY, NZ);
   clipper::Cell_descr cell_descr(NX, NY, NZ, 90, 90, 90);
   clipper::Cell cell(cell_descr);
   write_map(data_host, cell, grid, 1.0, "start.map");

   do_fft(data_host, cell, grid, "nxmap.map");

   delete [] data_host;
}

void do_molecule_map(const std::string &input_map_file_name) {

   // caller deletes data_host (not elegant)
   map_file_info_t grid_data_info = fill_data_from_map_file(input_map_file_name);
   const clipper::Grid &grid = grid_data_info.grid;
   cufftComplex *data_host = grid_data_info.data;
   const clipper::Cell &cell = grid_data_info.cell;
   // write_map(data_host, cell, grid, 1.0, "test-5dri-start.map"); is fine
   do_fft(data_host, cell, grid, "test-blur.map");

   delete [] data_host;
}


//  caller has filled data host, we make data_dev and transfer it around.
//
void do_fft(cufftComplex *data_host, const clipper::Cell &cell, const clipper::Grid &grid, 
            const std::string &output_map_file_name) {

   unsigned int nx, ny, nz;
   nx = grid.nu();
   ny = grid.nv();
   nz = grid.nw();
   std::cout << "DEBUG:: do_fft() nx ny nz " << nx << " " << ny << " " << nz << std::endl;

   std::cout << "Cuda Malloc" << std::endl;
   cufftComplex *data_dev;
   cudaMalloc(&data_dev, sizeof(cufftComplex)*nx*ny*nz);
   if (cudaGetLastError() != cudaSuccess){
      fprintf(stderr, "Cuda error: Failed to allocate\n");
           return;
   }

   // data -> device
   std::cout << "Cuda Memcpy" << std::endl;
   cudaMemcpy(data_dev, data_host, nx * ny * nz * sizeof(cufftComplex), cudaMemcpyHostToDevice);
   if (cudaGetLastError() != cudaSuccess){
      fprintf(stderr, "Cuda error: Failed cudaMemcpy()\n");
           return;
   }

   int nn[3] = {nx, ny, nz};

   /* Create a 3D FFT plan. */
   if (cufftPlanMany(&plan, 3, nn,
        NULL, 1, nx*ny*nz, // *inembed, istride, idist
        NULL, 1, nx*ny*nz, // *onembed, ostride, odist
        CUFFT_C2C, BATCH) != CUFFT_SUCCESS) {
        std::cerr << "CUFFT error: Plan creation failed" << std::endl;
        // need to free data_dev here
        return;
   }

   std::cout << "Cuda FFT Exec " << std::endl;
   auto tp_0 = std::chrono::high_resolution_clock::now();

   /* Use the CUFFT plan to transform the signal in place. */
   int n_times = 1;
   for (int i=0; i<n_times; i++) {
      if (cufftExecC2C(plan, data_dev, data_dev, CUFFT_FORWARD) != CUFFT_SUCCESS) {
         fprintf(stderr, "CUFFT error: ExecC2C Forward failed");
         return;
      }
   }

   if (cudaDeviceSynchronize() != cudaSuccess) {
      std::cout << "Cuda error: post cufftExecC2C() Failed to synchronize\n";
      return;
   } 
   auto tp_1 = std::chrono::high_resolution_clock::now();
   std::cout << "Cuda Synched" << std::endl;
   auto d10 = std::chrono::duration_cast<std::chrono::milliseconds>(tp_1 - tp_0).count();
   std::cout << "info:: " << d10 << " milliseconds for FFT call\n";


   // data -> host
   cudaMemcpy(data_host, data_dev, nx * ny * nz * sizeof(cufftComplex), cudaMemcpyDeviceToHost);

   // check the transformed data

   for (unsigned int i=0; i<nx; i++) {
      for (unsigned int j=0; j<ny; j++) {
         for (unsigned int k=0; k<nz; k++) {
            unsigned int idx = k + nz * j + nz * ny * i; // or the other way round?
            float reso = i*i + j*j + k+k;
            float re = data_host[idx].x;
            float im = data_host[idx].y;
            // if (idx %7 == 0)
            if (false)
               std::cout << "transformed-data " << i << " " << j << " " << k << " " << reso << " " << re << " " << im << std::endl;
         }
      }
   }

   // cut the fft - make this a kernel so that we don't have to move the data to the host
   // Better is to multiply by a gaussian (in a kernel).
   //
   for (unsigned int i=0; i<nx/2; i++) {
       for (unsigned int j=0; j<ny/2; j++) {
          for (unsigned int k=0; k<nz/2; k++) {
             unsigned int idx = k + nz * j + nz * ny * i; // or the other way round?
             float reso = i*i + j*j + k*k;
             // so that reso goes from ~0 (low res) to ~1 (high res).
             reso /= static_cast<float>(nx*nx);
             // if (reso > 16 * 16) {
             if (false) {
                data_host[idx].x = 0.0;
                data_host[idx].y = 0.0;
             }
             // blur
             float bf = exp(-reso*85.0);
             int qi = nx-1 -i;
             int qj = ny-1 -j;
             int qk = nz-1 -k;

             idx = k + nz * j + nz * ny * i;
             data_host[idx].x *= bf;
             data_host[idx].y *= bf;

             idx = k + nz * j + nz * ny * qi;
             data_host[idx].x *= bf;
             data_host[idx].y *= bf;

             idx = k + nz * qj + nz * ny * i;
             data_host[idx].x *= bf;
             data_host[idx].y *= bf;

             idx = k + nz * qj + nz * ny * qi;
             data_host[idx].x *= bf;
             data_host[idx].y *= bf;

             idx = qk + nz * j + nz * ny * i;
             data_host[idx].x *= bf;
             data_host[idx].y *= bf;

             idx = qk + nz * j + nz * ny * qi;
             data_host[idx].x *= bf;
             data_host[idx].y *= bf;

             idx = qk + nz * qj + nz * ny * i;
             data_host[idx].x *= bf;
             data_host[idx].y *= bf;

             idx = qk + nz * qj + nz * ny * qi;
             data_host[idx].x *= bf;
             data_host[idx].y *= bf;

             // data_host[idx].x = data_host[idx].x * data_host[idx].x + data_host[idx].y * data_host[idx].y;
             // data_host[idx].y = 0.0;
             // if (reso > 0.5) {
             if (i > NX/2 || j > NY/2 || k > NZ> 2) {
                data_host[idx].x = 0.0;
                data_host[idx].y = 0.0;
             }
          }
       }
   }

   // data -> device
   cudaMemcpy(data_dev, data_host, nx * ny * nz * sizeof(cufftComplex), cudaMemcpyHostToDevice);

   if (cufftExecC2C(plan, data_dev, data_dev, CUFFT_INVERSE) != CUFFT_SUCCESS){
      fprintf(stderr, "CUFFT error: ExecC2C Inverse failed");
      return;
   }

   /* 
    *  Results may not be immediately available so block device until all 
    *  tasks have completed
    */

   if (cudaDeviceSynchronize() != cudaSuccess) {
      fprintf(stderr, "CUDA error: Failed to synchronize\n");
      return;
   }

   // data -> host
   cudaMemcpy(data_host, data_dev, nx * ny * nz * sizeof(cufftComplex), cudaMemcpyDeviceToHost);

   /* 
    *  Divide by number of elements in data set to get back original data
    */

   float sf = 1.0/static_cast<float>(nx*ny*nz);
   for (unsigned int i=0; i<nx; i++) {
      for (unsigned int j=0; j<ny; j++) {
         for (unsigned int k=0; k<nz; k++) {
             unsigned int idx = k + nz * j + nz * ny * i; // or the other way round?
             float re = sf * data_host[idx].x;
             float im = sf * data_host[idx].y;
             float r2 = i*i + j*j + k*k;
             // std::cout << "back-transformed-data " << i << " " << j << " " << k << " " << r2 << " " << re << " " << im << std::endl;
         }
      }
   }

   write_map(data_host, cell, grid, sf, output_map_file_name);

   cufftDestroy(plan);
   cudaFree(data_dev);
}

void write_map(cufftComplex *data_host, const clipper::Cell &cell_in, const clipper::Grid &grid, float sf, const std::string &map_file_name) {

   // construct the nxmap
   unsigned int nu = grid.nu();
   unsigned int nv = grid.nv();
   unsigned int nw = grid.nw();

   std::vector<unsigned int> start_stop_crs(6,0);
   start_stop_crs[0] = 0;
   start_stop_crs[1] = nu-1;
   start_stop_crs[2] = 0;
   start_stop_crs[3] = nv-1;
   start_stop_crs[4] = 0;
   start_stop_crs[5] = nw-1;

   clipper::Grid_sampling clipper_grid_sampling(grid[0],
                                                grid[1],
                                                grid[2]);
   clipper::Coord_grid grid_min(start_stop_crs[0],
                                start_stop_crs[2],
                                start_stop_crs[4]);
   clipper::Coord_grid grid_max(start_stop_crs[1],
                                start_stop_crs[3],
                                start_stop_crs[5]);

    clipper::Grid_range clipper_grid_range(grid_min, grid_max);
    clipper::NXmap<float> nxmap;
    nxmap.init(cell_in, clipper_grid_sampling, clipper_grid_range);

    std::cout << "DEBUG:: write_map() cell " << cell_in.descr().format() << std::endl;
    std::cout << "DEBUG:: write_map() grid_sampling " << clipper_grid_sampling.format() << std::endl;
    std::cout << "DEBUG:: write_map() grid_range    " << clipper_grid_range.format()    << std::endl;
    int iu = 0;
    int iv = 0;
    int iw = 0;
    // int nu = start_stop_crs[1] - start_stop_crs[0];
    // int nv = start_stop_crs[3] - start_stop_crs[2];
    // int nw = start_stop_crs[5] - start_stop_crs[4];

   unsigned int count = 0;
   for (unsigned int i=0; i<nu; i++) {
      for (unsigned int j=0; j<nv; j++) {
         for (unsigned int k=0; k<nw; k++) {
             unsigned int idx = k + nw * j + nw * nv * i; // or the other way round?
             float re = sf * data_host[idx].x;
             float im = sf * data_host[idx].y;
             clipper::Coord_grid c_g(i,j,k);
             clipper::NXmap_base::Map_reference_coord mrc(nxmap, c_g);
             nxmap[mrc] = re;
             count++;
             if (false)
                if (idx%1000 == 0)
                   std::cout << std::setw(3) << i << " " << std::setw(3) << j << " " << std::setw(3) << k << " "
                             << std::setw(8) << std::fixed << std::setprecision(4) << std::right << re << std::endl;
         }
      }
   }

   std::cout << "DEBUG:: filled " << count << " grid points in write_map()" << std::endl;

      // write the map
      try { 
         clipper::CCP4MAPfile mapout;
         mapout.open_write(map_file_name);
         mapout.set_cell(cell_in);
         mapout.export_nxmap(nxmap);
         mapout.close_write();
      }
      catch (const clipper::Message_base &except) {
        std::cout << "WARNING:: Failed to write " << map_file_name  << std::endl;
      }
}

// Fill data (host), caller deletes data_host
//
map_file_info_t
fill_data_from_map_file(const std::string &map_file_name) {

   clipper::CCP4MAPfile file;
   clipper::NXmap<float> nxmap;
   file.open_read(map_file_name);

   std::cout << "=================== EM Map NXmap =================== " << std::endl;
   file.import_nxmap(nxmap);
   clipper::Grid grid = nxmap.grid();
   std::cout << "INFO:: created NX Map with grid " << nxmap.grid().format() << std::endl;

   // axis order is correct?

   unsigned int nx = nxmap.grid().nu();
   unsigned int ny = nxmap.grid().nv();
   unsigned int nz = nxmap.grid().nw();
   std::cout << "DEBUG:: fill_data_from_map_file() new for data_host "
              << nx << " " << ny << " " << nz << std::endl;
   cufftComplex *data_host = new cufftComplex[nx*ny*nz];

   clipper::NXmap_base::Map_reference_index ix;
   unsigned int idx = 0;
   for (ix = nxmap.first(); !ix.last(); ix.next() )  { // iterator index.
       float re = nxmap[ix];
       data_host[idx].x = re;
       data_host[idx].y = 0.0;
       idx++;
   }
   std::cout << "DEBUG:: fill_data_from_map_file() filled " << idx << " output grid points" << std::endl;

   map_file_info_t mfi(grid, file.cell(), data_host);
   return mfi;

}


