
#ifndef CUDA_BONDS_HH
#define CUDA_BONDS_HH

#include "CudaLBFGS/lbfgs.h"

#include "cuda-restraints.cuh"
#include "nxmap-cuda.cuh"

class gpu_bonds : public cost_function {

   unsigned int n_atoms;  // filled by constructor

   // cuda_restraints_transfer_container_t convert_to_cuda_restraints(const coot::restraints_container_t &restraints) const;

   // fill d_restraints_container_p and all its parts
   void transfer_restraints_to_device();

   float calculate_bond_distortion(const std::vector<cuda_bond_restraint_t> &restraints, float *x) const;
   cuda_restraints_container_t restraints_container; // contains device pointers

   int n_f_gradf; // how many times was the kernel called?

   unsigned int n_threads_per_block;
   unsigned int n_blocks;
   float map_weight;
   float lj_epsilon;

public:

   // non generic
   // gpu_bonds(const coot::restraints_container_t &rc, nxmap_cuda d_nxm);

   gpu_bonds(const cuda_restraints_transfer_container_t &rc, nxmap_cuda d_nxm, unsigned int n_atoms);

   // pass: d_x: pointer to an array of "coordinates" (model),
   //
   // returns: d_f: target function value for these coordinates
   //          d_grad: gradients for these coordinates
   //
   // a virtual function, I guess
   void f_gradf(const float *d_x, float *d_f, float *d_grad);

   ~gpu_bonds(); // clear up d_restraints_p;

   void set_launch_parameter(unsigned int n_blocks_in, unsigned int n_threads_per_block_in) {
      n_blocks = n_blocks_in;
      n_threads_per_block = n_threads_per_block_in;
   }

   void set_map_weight(const float &w) {
      map_weight = w;
   }

   void set_lj_epsilon(const float &e);

   // for debugging distortions (etc)
   cuda_restraints_transfer_container_t transfer_restraints;

   bool do_bonds;
   bool do_angles;
   bool do_torsions;
   bool do_planes;
   bool do_chiral_volumes;
   bool do_nbcs;
   bool do_geman_mcclure_distances;
   bool do_electron_density;

   nxmap_cuda nxm;
   float *d_distortion_score_electron_density_p; // device holder for ED sum
   int get_n_f_gradf() const { return n_f_gradf; }
   void print_refinement_flags() const;
   float get_map_weight() const { return map_weight; }

};


#endif // CUDA_BONDS_HH

