// -*-c++-*-
#include "clipper-cuda.cuh"

class nxmap_cuda {
 
     float *list_p;
 
     unsigned int list_size;
     unsigned int nu_, nv_, nw_;
     rt_cuda rt_orth_grid;
     rt_cuda rt_grid_orth;
     bool on_host;
 
     __device__ int nearest_int_cuda(const float &f) const {
        int ii = __float2int_rn(f);
        return ii;
     }

  public:
 
     nxmap_cuda(const clipper::NXmap<float> &nxmap, bool on_device=true);
 
     nxmap_cuda(const grid_cuda &grid, const rt_cuda &rtop_orth_grid, bool on_device=true);

     nxmap_cuda(const std::string &file_name); // fill with "stand-alone" data - use host memory

      nxmap_cuda(const nxmap_cuda &h_map, bool tr) {
         import(h_map);
      }
 
     nxmap_cuda() {
        list_p = NULL;
        list_size = 0;
        nu_ = 0; nv_ = 0; nw_ = 0;
        bool on_host = true;
     }

   /*
   __host__ __device__ ~nxmap_cuda() {
      if (on_host) {
	 printf("~nxmap_cuda list_p is 0%p\n", list_p);
	 printf("deleting host memory for map...\n");
	 // fflush(stdout);
	 delete [] list_p;
	 list_p = 0;
	 printf("done delete host memory for map\n");
	 // fflush(stdout);
      }
      }
   */

   // don't use a destructor, use a clear function - or a shared pointer later
   void clear() {
      if (on_host)
	 delete [] list_p;
      else
	 cudaFree(list_p);
   }

/* -------- crash -----------
     __host__ __device__ ~nxmap_cuda() {
	 if (on_host)
	    delete list_p;
	 else
	    cudaFree(list_p);
     }
*/
     __host__ void write_map(const std::string &file_name) const;
 
     grid_cuda grid;
 
     __host__ __device__ float& at(const coord_grid_cuda &cg) {
        const int &iu = cg.u;
        const int &iv = cg.v;
        const int &iw = cg.w;
        int idx =  iw + iv * nw_ + iu * nw_ * nv_;
        if (idx >= list_size) {
           printf("error in grid index access attempt %d %d %d vs grid %d %d %d\n", iu, iv, iw, grid.nu, grid.nv, grid.nw);
           return list_p[0];
        } else {
           return list_p[idx];
        }
     }
 
     __host__ __device__ void set_data(const coord_grid_cuda &cg, const float &f) {
        const int &iu = cg.u;
        const int &iv = cg.v;
        const int &iw = cg.w;
        int idx =  iw + iv * nw_ + iu * nw_ * nv_;
        if (idx >= list_size) {
           printf("error in grid index access attempt %d %d %d vs grid %d %d %d\n", iu, iv, iw, grid.nu, grid.nv, grid.nw);
        } else {
           list_p[idx] = f;
        }
     }
 
     rt_cuda operator_orth_grid() const { return rt_orth_grid; }
     rt_cuda operator_grid_orth() const { return rt_grid_orth; }
 
     __host__ __device__ unsigned int nu() const { return nu_; }
     __host__ __device__ unsigned int nv() const { return nv_; }
     __host__ __device__ unsigned int nw() const { return nw_; }
 
     void import(const nxmap_cuda &h_map);

     __device__ coord_grid_cuda get_coord_grid(unsigned int &idx_in) const {
        unsigned int iu = idx_in/(nv()*nw());
        unsigned int r = idx_in - iu*nv()*nw();
        unsigned int iv = r/nw();
        r -= iv*nw();
        return coord_grid_cuda(iu, iv, r);
     }
 
     __device__ coord_grid_cuda coord_grid(const coord_orth_cuda &co) const {
        coord_map_cuda cm = coord_map(co);
        coord_grid_cuda cg = coord_grid(cm);
        printf("debug in coord_grid() %f %f %f %d %d %d\n", cm.u(), cm.v(), cm.w(), cg.u, cg.v, cg.w);
        return cg;
     }
 
     __device__ coord_map_cuda coord_map(const coord_orth_cuda &co) const {
        printf("debug in coord_map(): rt_orth_grid:\n");
        rt_orth_grid.print_self();
        return rt_orth_grid * co;
     }
 
     __device__ coord_map_cuda coord_map(const coord_grid_cuda &cg) const {
        return coord_map_cuda(static_cast<float>(cg.u), static_cast<float>(cg.v), static_cast<float>(cg.w));
     }
 
     __device__ coord_grid_cuda coord_grid(const coord_map_cuda &cm) const {
 
       int iu = nearest_int_cuda(cm.u());
       int iv = nearest_int_cuda(cm.v());
       int iw = nearest_int_cuda(cm.w());
       return coord_grid_cuda(iu, iv, iw);
     }
 
     __device__ float interp_nearest(const coord_map_cuda &c_m) const {
       int iu = nearest_int_cuda(c_m.u());
       int iv = nearest_int_cuda(c_m.v());
       int iw = nearest_int_cuda(c_m.w());
       int idx =  iw + iv* nw_ + iu * nw_ * nv_;
       // printf("idx: %d\n", idx);
       float f = list_p[idx];
       return f;
     }
 
     __host__ __device__ float density_at_grid_point(const coord_grid_cuda &c_g) const {
       const int &iu = c_g.u;
       const int &iv = c_g.v;
       const int &iw = c_g.w;
       int idx =  iw + iv * nw_ + iu * nw_ * nv_;
       float f = list_p[idx];
       return f;
     }
 
     // I guess I need a in_map() test
     __host__ __device__ bool in_map(const coord_grid_cuda &cg) const {
        if (cg.u < 0) {
           return false;
        } else {
           if (cg.v < 0) {
              return false;
           } else {
              if (cg.w < 0) {
                 return false;
              } else {
                 if (cg.u >= nu()) {
                    return false;
                 } else {
                    if (cg.v >= nu()) {
                       return false;
                    } else {
                       if (cg.w >= nu()) {
                          return false;
                       }
                    }
                 }
              }
           }
        }
        return true;
     }
 
     __host__ __device__ unsigned int size() const { return list_size; }
 
     // co is in orth coordinates
     __device__ float density_at_point_interp_nearest(const coord_orth_cuda &co) const {
        // make the grid point
        coord_map_cuda c_m(rt_orth_grid * co);
        // printf("c_m: %f %f %f\n", c_m.x, c_m.y, c_m.z);
        return interp_nearest(c_m);
     }
 
     __device__ bool can_interp(const coord_map_cuda &cm) const {
         coord_grid_cuda cg = coord_grid(cm.floor());
         coord_grid_cuda c_minus = cg + coord_grid_cuda(-1, -1, -1);
         if (! in_map(c_minus)) {
            return false;
         } else {
            coord_grid_cuda cg_plus = cg + coord_grid_cuda(1,1,1);
            return in_map(cg_plus);
         }
     }
 
     // co is in map space (a coord_map)
     __device__ float density_at_point_interp_linear(const coord_map_cuda &cm) const  {
 
        // based off of map_interp linear from clipper.
 
        if (! can_interp(cm)) {
           // not sure what to do here at the moment.
           return 0.0f;
        } else {
           float u0 = floor(cm.u());
           float v0 = floor(cm.v());
           float w0 = floor(cm.w());
 
           // a 3-int coordinate
           coord_grid_cuda r(static_cast<int>(u0), static_cast<int>(v0), static_cast<int>(w0));
 
           float cu1 = cm.u() - u0;
           float cv1 = cm.v() - v0;
           float cw1 = cm.w() - w0;
           float cu0 = 1.0 - cu1;
           float cv0 = 1.0 - cv1;
           float cw0 = 1.0 - cw1;
 
           float r00 = cw0 * density_at_grid_point(r);
           r00 += cw1 * density_at_grid_point(r.next_w());
           float r01 = cw1 * density_at_grid_point(r.next_v());
           r01 += cw0 * density_at_grid_point(r.prev_w());
           float r11 = cw0 * density_at_grid_point(r.next_u());
           r11 += cw1 * density_at_grid_point(r.next_w());
           float r10 = cw1 * density_at_grid_point(r.prev_v());
           r10 += cw0 * density_at_grid_point(r.prev_w());
           float val = ( cu0*( cv0*r00 + cv1*r01 ) + cu1*( cv0*r10 + cv1*r11 ) );
           return val;
        }
     }
 
     __device__ float density_at_point_interp_cubic(const coord_orth_cuda &co) const {
 
        coord_map_cuda c_m(rt_orth_grid * co);
        return density_at_point_interp_cubic(c_m);
 
     }
 
     // Put this in its own file - clipper copyright
     //
     __device__ float density_at_point_interp_cubic(const coord_map_cuda &pos) const {

       if (! can_interp(pos)) {
           // not sure what to do here at the moment.
	  return 0.0f;
       } else {
	  float u0 = floor( pos.u() );
	  float v0 = floor( pos.v() );
	  float w0 = floor( pos.w() );

	  coord_grid_cuda iw, iv;
	  coord_grid_cuda iu(static_cast<int>(u0)-1, static_cast<int>(v0)-1, static_cast<int>(w0)-1);

	  float su, sv, sw, cu[4], cv[4], cw[4];
	  float cu1( pos.u() - u0 );
	  float cv1( pos.v() - v0 );
	  float cw1( pos.w() - w0 );
	  float cu0( 1.0f - cu1 );
	  float cv0( 1.0f - cv1 );
	  float cw0( 1.0f - cw1 );

	  cu[0] = -0.5*cu1*cu0*cu0; // cubic spline coeffs: u
	  cu[1] = cu0*( -1.5*cu1*cu1 + cu1 + 1.0 );
	  cu[2] = cu1*( -1.5*cu0*cu0 + cu0 + 1.0 );
	  cu[3] = -0.5*cu1*cu1*cu0;
	  cv[0] = -0.5*cv1*cv0*cv0; // cubic spline coeffs: v
	  cv[1] = cv0*( -1.5*cv1*cv1 + cv1 + 1.0 );
	  cv[2] = cv1*( -1.5*cv0*cv0 + cv0 + 1.0 );
	  cv[3] = -0.5*cv1*cv1*cv0;
	  cw[0] = -0.5*cw1*cw0*cw0; // cubic spline coeffs: w
	  cw[1] = cw0*( -1.5*cw1*cw1 + cw1 + 1.0 );
	  cw[2] = cw1*( -1.5*cw0*cw0 + cw0 + 1.0 );
	  cw[3] = -0.5*cw1*cw1*cw0;
	  su = 0.0;
	  int i, j;
	  for ( j = 0; j < 4; j++ ) {
	     iv = iu;
	     sv = 0.0;
	     for ( i = 0; i < 4; i++ ) {
		iw = iv;
		sw  = cw[0] * density_at_grid_point(iw);
		sw += cw[1] * density_at_grid_point(iw.next_w());
		sw += cw[2] * density_at_grid_point(iw.next_w());
		sw += cw[3] * density_at_grid_point(iw.next_w());
		sv += cv[i] * sw;
		iv.next_v();
	     }
	     su += cu[j] * sv;
	     iu.next_u();
	  }
	  return su;
       }
     }


   __device__ grad_map_cuda grad_grid(const coord_grid_cuda &r) const {
        return grad_map_cuda(0.5*(density_at_grid_point(coord_grid_cuda(r.u+1, r.v, r.w))-density_at_grid_point(coord_grid_cuda(r.u-1, r.v, r.w))),
                             0.5*(density_at_grid_point(coord_grid_cuda(r.u, r.v+1, r.w))-density_at_grid_point(coord_grid_cuda(r.u, r.v-1, r.w))),
                             0.5*(density_at_grid_point(coord_grid_cuda(r.u, r.v, r.w+1))-density_at_grid_point(coord_grid_cuda(r.u, r.v, r.w-1))));
   }

     // co is in coord orth space
     __device__ grad_orth_cuda grad_interp_linear(const coord_orth_cuda &co) const  {
	coord_map_cuda cm = rt_orth_grid * co;
	if (! can_interp(cm)) {
	   return grad_map_cuda(0.0f, 0.0f, 0.0f);
	} else {
	   // printf("grad_interp_linear(): can interp: %f %f %f\n", cm.u(), cm.v(), cm.z());
	   grad_map_cuda gm = grad_interp_linear(cm);
	   grad_orth_cuda go(rt_grid_orth.rot * gm);
	   return go;
	}
     }


   // coord_map is in map space
   __device__ grad_map_cuda grad_interp_linear(const coord_map_cuda &coord_map) const  {

      // based off of map_interp linear from clipper.

      float u0 = floor(coord_map.u());
      float v0 = floor(coord_map.v());
      float w0 = floor(coord_map.w());

      // a 3-int coordinate
      coord_grid_cuda r(static_cast<int>(u0), static_cast<int>(v0), static_cast<int>(w0));

      float cu1 = coord_map.u() - u0;
      float cv1 = coord_map.v() - v0;
      float cw1 = coord_map.w() - w0;
      float cu0 = 1.0 - cu1;
      float cv0 = 1.0 - cv1;
      float cw0 = 1.0 - cw1;

      if (false) {
         printf("grad_interp_linear grad_grid        %d %d %d\n", r.u, r.v, r.w);
         printf("grad_interp_linear grad_grid next_w %d %d %d\n", r.next_w().u, r.next_w().v, r.next_w().w);
         printf("grad_interp_linear grad_grid prev_w %d %d %d\n", r.prev_w().u, r.prev_w().v, r.prev_w().w);
      }

      grad_map_cuda r00 = grad_grid(r) * cw0;
      r00  += grad_grid(r.next_w()) * cw1;
      grad_map_cuda r01 = grad_grid(r.next_v()) * cw1;
      r01  += grad_grid(r.prev_w()) * cw0;
      grad_map_cuda r11 = grad_grid(r.next_u()) * cw0;
      r11  += grad_grid(r.next_w()) * cw1;
      grad_map_cuda r10 = grad_grid(r.prev_v()) * cw1;
      r10  += grad_grid(r.prev_w()) * cw0;
      grad_map_cuda val = ( ( r00*cv0 + r01*cv1 )  * cu0 + ( r10*cv0 + r11*cv1 )  * cu1);
      return val;
   }

     // co is in coord orth space
     __device__ grad_orth_cuda grad_interp_cubic(const coord_orth_cuda &co) const  {
	coord_map_cuda cm = rt_orth_grid * co;
	if (! can_interp(cm)) {
	   return grad_map_cuda(0.0f, 0.0f, 0.0f);
	} else {
	   grad_map_cuda gm = grad_interp_cubic(cm);
	   grad_orth_cuda go(rt_grid_orth.rot * gm);
	   return go;
	}
     }

     // Put this in its own file clipper copyright
     //
     __device__ grad_map_cuda grad_interp_cubic(const coord_map_cuda &pos) const {
	float u0 = floor( pos.u() );
	float v0 = floor( pos.v() );
	float w0 = floor( pos.w() );
	// typename M::Map_reference_coord iw, iv, iu( map, Coord_grid( int(u0)-1, int(v0)-1, int(w0)-1 ) );
	coord_grid_cuda iw, iv;
	coord_grid_cuda iu(int(u0)-1, int(v0)-1, int(w0)-1);

	float s1, s2, s3, du1, dv1, dv2, dw1, dw2, dw3;
	float cu[4], cv[4], cw[4], gu[4], gv[4], gw[4];
	float cu1( pos.u() - u0 );
	float cv1( pos.v() - v0 );
	float cw1( pos.w() - w0 );
	float cu0( 1.0 - cu1 );
	float cv0( 1.0 - cv1 );
	float cw0( 1.0 - cw1 );

	cu[0] = -0.5*cu1*cu0*cu0; // cubic spline coeffs: u
	cu[1] = cu0*( -1.5*cu1*cu1 + cu1 + 1.0 );
	cu[2] = cu1*( -1.5*cu0*cu0 + cu0 + 1.0 );
	cu[3] = -0.5*cu1*cu1*cu0;
	cv[0] = -0.5*cv1*cv0*cv0; // cubic spline coeffs: v
	cv[1] = cv0*( -1.5*cv1*cv1 + cv1 + 1.0 );
	cv[2] = cv1*( -1.5*cv0*cv0 + cv0 + 1.0 );
	cv[3] = -0.5*cv1*cv1*cv0;
	cw[0] = -0.5*cw1*cw0*cw0; // cubic spline coeffs: w
	cw[1] = cw0*( -1.5*cw1*cw1 + cw1 + 1.0 );
	cw[2] = cw1*( -1.5*cw0*cw0 + cw0 + 1.0 );
	cw[3] = -0.5*cw1*cw1*cw0;
	gu[0] =  cu0*( 1.5*cu1 - 0.5 ); // cubic spline grad coeffs: u
	gu[1] =  cu1*( 4.5*cu1 - 5.0 );
	gu[2] = -cu0*( 4.5*cu0 - 5.0 );
	gu[3] = -cu1*( 1.5*cu0 - 0.5 );
	gv[0] =  cv0*( 1.5*cv1 - 0.5 ); // cubic spline grad coeffs: v
	gv[1] =  cv1*( 4.5*cv1 - 5.0 );
	gv[2] = -cv0*( 4.5*cv0 - 5.0 );
	gv[3] = -cv1*( 1.5*cv0 - 0.5 );
	gw[0] =  cw0*( 1.5*cw1 - 0.5 ); // cubic spline grad coeffs: w
	gw[1] =  cw1*( 4.5*cw1 - 5.0 );
	gw[2] = -cw0*( 4.5*cw0 - 5.0 );
	gw[3] = -cw1*( 1.5*cw0 - 0.5 );
	s1 = du1 = dv1 = dw1 = 0.0;

	for (int j = 0; j < 4; j++ ) {
	   iv = iu;
	   s2 = dv2 = dw2 = 0.0;
	   for (int i = 0; i < 4; i++ ) {
	      iw = iv;
	      s3  = cw[0] * density_at_grid_point(iw);
	      dw3 = gw[0] * density_at_grid_point(iw);
	      iw.next_w();
	      s3  += cw[1] * density_at_grid_point(iw);
	      dw3 += gw[1] * density_at_grid_point(iw);
	      iw.next_w();
	      s3  += cw[2] * density_at_grid_point(iw);
	      dw3 += gw[2] * density_at_grid_point(iw);
	      iw.next_w();
	      s3  += cw[3] * density_at_grid_point(iw);
	      dw3 += gw[3] * density_at_grid_point(iw);
	      s2  += cv[i] * s3;
	      dv2 += gv[i] * s3;
	      dw2 += cv[i] * dw3;
	      iv.next_v();
	   }
	   s1 +=  cu[j] * s2;
	   du1 += gu[j] * s2;
	   dv1 += cu[j] * dv2;
	   dw1 += cu[j] * dw2;
	   iu.next_u();
	}
	grad_map_cuda grad(du1, dv1, dw1);
	return grad;
     }

   // co is in (normal molecular coordinates) orth space
   __device__ float density_at_point_interp_linear(const coord_orth_cuda &co) const  {
      coord_map_cuda c_m(rt_orth_grid * co);
      // printf("density_at_point_interp_linear %f %f %f grid: %f %f %f\n", co.x(), co.y(), co.z(), c_m.u(), c_m.v(), c_m.w());
      return density_at_point_interp_linear(c_m);
  }

   __device__ float density_at_point(const float &x, const float &y, const float &z) const {
      coord_orth_cuda co(x,y,z);
      return density_at_point_interp_nearest(co);
   }

   __device__ void find_peak() const {
      float best = -999.9;
      for (unsigned int i=0; i<list_size; i++) { 
          float f = list_p[i];
          if (f > best) {
              best = f;
              printf("new best %d %f\n", i, f);
          }
      }

      printf("------\n");
      best = -999.9;
      for(unsigned int iu=0; iu<nu_; iu++) {
         for(unsigned int iv=0; iv<nv_; iv++) {
            for (unsigned int iw=0; iw<nw_; iw++) {
	       int idx =  iw + iv* nw_ + iu * nw_ * nv_;
	       float f = list_p[idx];
	       if (f > best) {
		  best = f;
		  printf("new best %d %f\n", idx, f);
	       }
            }
         }
      }
   }
 
};

